import os

import inflatox
from inflatox.consistency_conditions import GeneralisedAL
import numpy as np
import sympy
import pytest

a = 1 / 600
m_phi = 2e-5
m_chi = m_phi * np.sqrt(9)
args = np.array([a, m_chi, m_phi])


@pytest.fixture
def angular_model():
    model = "angular"

    # setup the coordinates
    p, x = sympy.symbols("phi chi")
    coords = [p, x]

    # setup the potential
    mp, mx, a = sympy.symbols("m_phi m_chi alpha")
    potential = a / 2 * ((mp * p) ** 2 + (mx * x) ** 2).nsimplify()

    # setup the metric
    metric_diagonal = 6 * a / (1 - p**2 - x**2) ** 2
    metric = [[0 for _ in range(2)] for _ in range(2)]
    metric[0][0] = metric_diagonal
    metric[1][1] = metric_diagonal

    hesse = inflatox.InflationModelBuilder.new(
        coords,
        metric,
        potential,
        model_name=model,
        silent=True,
    ).build()

    out = inflatox.Compiler(hesse).compile()
    return GeneralisedAL(out)


def test_hesse(angular_model):
    extent = (-1.05, 1.05, -1.05, 1.05)
    N = [100, 100]
    print(angular_model.calc_H(np.array([0.0, 0.0]), args))
    angular_model.calc_H_array(args, *extent, N)


def test_potential(angular_model):
    extent = [(-1.05, 1.05), (-1.05, 1.05)]
    N = [100, 100]
    print(angular_model.calc_V(np.array([0.0, 0.0]), args))
    angular_model.calc_V_array(args, *extent, N)



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

import pytest
import sympy
from inflatox.compiler import *


@pytest.fixture
def cprinter():
    x, y, a, b, xdot, ydot = sympy.symbols("x y a b \\dot{{x}} \\dot{{y}}")
    return CInflatoxPrinter([x, y], [xdot, ydot])


@pytest.fixture
def gslprinter():
    x, y, a, b, xdot, ydot = sympy.symbols("x y a b \\dot{{x}} \\dot{{y}}")
    return GSLInflatoxPrinter([x, y], [xdot, ydot])


def test_cinflatox_printer(cprinter):
    x, y, a, b, xdot, ydot = sympy.symbols("x y a b \\dot{{x}} \\dot{{y}}")
    # test basic symbols
    assert "x[0]" == cprinter._print_Symbol(x)
    assert "x[1]" == cprinter._print_Symbol(y)
    assert "args[0]" == cprinter._print_Symbol(a)
    assert "args[1]" == cprinter._print_Symbol(b)
    assert "xdot[0]" == cprinter._print_Symbol(xdot)
    assert "xdot[1]" == cprinter._print_Symbol(ydot)

    # test basic arithmetic
    assert "pow(x[0], 2) + x[1]" == cprinter.doprint(x**2 + y)
    assert "x[0]*x[1]" == cprinter.doprint(x * y)

    # test some math.h funcs
    assert "sqrt(args[0])*x[1]" == cprinter.doprint(sympy.sqrt(a) * y)
    assert "sin(x[0])" == cprinter.doprint(sympy.sin(x))


def test_gslinflatox_printer_headers(gslprinter):
    x, y = sympy.symbols("x y")
    gslprinter.doprint(sympy.besselj(1, x))
    assert gslprinter.BESSELH in gslprinter.required_headers
    gslprinter.doprint(sympy.hyper([], [1], x))
    assert gslprinter.HYPERH in gslprinter.required_headers


def test_gslinflatox_bessel(gslprinter):
    x, y, n = sympy.symbols("x y n")
    assert "gsl_sf_bessel_J0(x[0])" == gslprinter.doprint(sympy.besselj(0, x))
    assert "gsl_sf_bessel_J1(x[0])" == gslprinter.doprint(sympy.besselj(1, x))
    assert "gsl_sf_bessel_Jn(10, x[0])" == gslprinter.doprint(sympy.besselj(10, x))
    assert "gsl_sf_bessel_Jnu(0.50000000000000000, x[0])" == gslprinter.doprint(
        sympy.besselj(0.5, x)
    )


def test_gslinflatox_hyper(gslprinter):
    x, y, n = sympy.symbols("x y n")
    assert "gsl_sf_hyperg_2F0(0, 1, x[0])" == gslprinter.doprint(sympy.hyper([0, 1], [], x))
    assert "gsl_sf_hyperg_2F1(0, 1, 2, x[0])" == gslprinter.doprint(sympy.hyper([0, 1], [2], x))
    assert "gsl_sf_hyperg_1F1(0, 1, x[0])" == gslprinter.doprint(sympy.hyper([0], [1], x))
    assert "gsl_sf_hyperg_0F1(0, x[0])" == gslprinter.doprint(sympy.hyper([], [0], x))
    with pytest.raises(Exception) as exinfo:
        gslprinter.doprint(sympy.hyper([0, 3, 4], [1, 2], x))
    assert "Cannot compute" in str(exinfo.value)



#  Copyright© 2023 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

import numpy as np
import sympy
from interruptingcow import timeout
import os
from . import libinflx_rs
from joblib import Parallel, cpu_count, delayed
from sympy.simplify import sqrtdenest
from sympy.vector import Gradient


class InflationModel:
    """This class describes a multi-field inflation model. It contains all the symbolic expressions
    needed by inflatox to compute consistency conditions, solve the equations of motion etc.

    To obtain an instance of this class, you can either:
    1. Construct it directly
        This requires manually computing a fair few expressions, which can result in much smaller
        and thus more efficient expressions.
    2. Construct it automatically using the `SymbolicCalculation` class
        Using the `SymbolicCalculation` class is much easier: you only need to provide said class
        with an expression for the field-space metric and potential and it will automatically
        derive all other quantities. The disadvantage of that approach is that it may produce
        suboptimal expressions.
    """

    def __init__(
        self,
        model_name: str,
        coordinates: list[sympy.Symbol],
        tangents: list[sympy.Symbol],
        basis: list[list[sympy.Expr]],
        eom_fields: list[sympy.Expr],
        eom_h: sympy.Expr,
        eom_hdot: sympy.Expr,
        potential: sympy.Expr,
        metric: list[list[sympy.Expr]],
        gradient_square: sympy.Expr,
        hesse_cmp: list[list[sympy.Expr]],
    ):
        self.model_name = model_name
        self.hesse_cmp = hesse_cmp
        self.coordinates = coordinates
        self.coordinate_tangents = tangents
        self.dim = len(coordinates)
        self.basis = basis
        self.eom_fields = eom_fields
        self.eom_h = eom_h
        self.eom_hdot = eom_hdot
        self.potential = potential
        self.metric = metric
        self.gradient_square = gradient_square
        if len(hesse_cmp[0]) != len(hesse_cmp):
            raise Exception(
                "The Hesse matrix is square; the provided list was not (number of columns != number of rows)"
            )
        if len(metric[0]) != len(metric):
            raise Exception(
                "The metric tensor is square; the provided list was not (number of columns != number of rows)"
            )
        if len(hesse_cmp[0]) != len(basis[0]):
            raise Exception("The provided Hesse Matrix and basis are of different dimensionality")
        if len(basis) != len(coordinates):
            raise Exception(
                "The dimension of the provided basis does not match the number of fields."
            )
        if len(coordinates) != len(tangents):
            raise Exception(
                "The number of coordinate symbols does not match the number of tangent symbols."
            )

    def __str__(self):
        return f"""[Inflatox Inflation Model]
model name: {self.model_name}
dimensionality: {self.dim} field(s)
coordinates: {[f for f in self.coordinates]}
potential: {self.potential}
metric: {sympy.Matrix(self.metric)}
basis vectors (cntr. var.): {[sympy.Matrix(vec) for vec in self.basis]}
hesse matrix: {sympy.Matrix(self.hesse_cmp)}
    """


class SimplificationTimeOut(Exception):
    """This exception is thrown when a simplification exceeds a user-defined time-out time. This
    exception is meant to be caught."""

    pass


class InflationModelBuilder:
    """This class is a helper-class for producing instances of `InflationModel`. After obtaining
    an instance of this class, an instance of `InflationModel` can be obtained with the `.execute()`
    method.

    ### Usage
    An instance of this `SymbolicCalculation` class may be constructed from a
    user-defined metric on the scalar manifold, a potential and a list of sympy
    Symbols indicating which symbols (used in the definition of the potential and
    metric) should be interpreted as fields.

    Calling the `.execute()` method on an instance of `SymbolicCalculation` with
    an appropriate starting point for constructing the vielbein basis will perform
    the calculation of the components of the covariant Hesse matrix projected
    onto the vielbein.
    """

    @classmethod
    def new(
        cls,
        fields: list[sympy.Symbol],
        field_metric: list[list[sympy.Symbol]],
        potential: sympy.Expr,
        model_name: str | None = None,
        silent: bool = False,
        init_sympy_printing: bool = True,
        assertions: bool = True,
        simplify: bool = True,
        simplify_timeout: float | None = None,
    ):
        """Constructor for `SymbolicCalculation`

        ### Note
        This class will automatically derive names for the derivatives of the fields (these are
        used by other inflatox components when solving the equations of motion).

        Simplifications are currently disabled on windows due to their time-out system relying on
        UNIX signals.

        ### Args
        - `fields` (`list[sympy.Symbol]`): list of sympy symbols that should be
          interpreted as fields (coordinates on the scalar manifold).
        - `field_metric` (`list[list[sympy.Symbol]]`): metric tensor on the scalar
          manifold.
        - `potential` (`sympy.Expr`): potential of the scalar field theory.
        - `model_name` (`str | None`, *optional*): name of the model (potential +
          scalar manifold geometry). Used for printing. Defaults to None.
        - `silent` (`bool`, *optional*): if True, no console output will be produced
          during the calculation. Defaults to False.
        - `init_sympy_printing` (`bool`, *optional*): if True, pretty printing of sympy
          expressions will be initialised with default options. Set to False if you
          want to use your own settings for printing. Defaults to True.
        - `assertions` (`bool`, *optional*): if False, expensive intermediate
          assertions will be disabled. Defaults to False.
        - `simplify` (`bool`, *optional*): When `True` `sympy`'s simplify method will be used.
          Due to the time-out system for simplifications relying on UNIX signals, simplifications
          cannot be turned on on Windows. If you set `simplify` to `True` on Windows it will be
          ignored. Defaults to `True` (except on Windows platforms).
        - `simplify_timeout` (`float`, *optional*): time-out time in seconds for simplification
          steps.

        ### Returns
        `SymbolicCalculation`: object that can be used to perform the symbolic
        calculation by calling the `.execute()` method.
        """
        if init_sympy_printing:
            sympy.init_printing()

        if simplify == True and os.name == "nt":
            libinflx_rs.log_warn(
                "cannot use simplifications on Windows. Continuing without simplifications."
            )
            return cls(
                fields=fields,
                field_metric=field_metric,
                potential=potential,
                model_name=model_name if model_name is not None else "generic model",
                silent=silent,
                assertions=assertions,
                simplify=False,
                simplify_timeout=0.0,
            )

        return cls(
            fields=fields,
            field_metric=field_metric,
            potential=potential,
            model_name=model_name if model_name is not None else "generic model",
            silent=silent,
            assertions=assertions,
            simplify=simplify,
            simplify_timeout=simplify_timeout if simplify_timeout is not None else 20.0,
        )

    def __init__(
        self,
        fields: list[sympy.Symbol],
        field_metric: list[list[sympy.Symbol]],
        potential: sympy.Expr,
        model_name: str,
        silent: bool,
        assertions: bool,
        simplify: bool,
        simplify_timeout: float,
    ):
        """Internal constructor"""
        assert len(field_metric) == len(field_metric[0]), "field metric should be square"
        assert len(field_metric) == len(fields), (
            "number of fields must match dimensionality of metric tensor"
        )

        self.model_name = model_name
        self.dim = len(fields)
        self.fields = fields
        self.field_derivatives = sympy.symbols([f"\\dot{{{sympy.latex(phi)}}}" for phi in fields])
        self.metric = field_metric
        self.V = potential
        self.assertions = assertions
        self.silent = silent
        self.simplify = simplify
        self.simplify_timeout = simplify_timeout

    def simplify_expr(self, expr: sympy.Expr) -> sympy.Expr:
        """simplifies expression"""
        if not self.simplify or os.name == "nt":
            return expr
        try:
            with timeout(self.simplify_timeout, exception=SimplificationTimeOut):
                return sympy.simplify(expr, ratio=1, inverse=True)
        except SimplificationTimeOut:
            print(f"""Simplification step timed out (>{self.simplify_timeout}s)!
Consider increasing the simpliciation time-out time or turning off simplifications.""")
            return expr

    def expand_and_factor_expr(self, expr: sympy.Expr) -> sympy.Expr:
        """`sympy.expand` followed by `sympy.factor` with a time-out"""
        if not self.simplify or os.name == "nt":
            return expr
        try:
            with timeout(self.simplify_timeout, exception=SimplificationTimeOut):
                return sympy.factor(sympy.expand(expr))
        except SimplificationTimeOut:
            print(f"""Simplification step timed out (>{self.simplify_timeout}s)!
Consider increasing the simpliciation time-out time or turning off simplifications.""")
            return expr

    def sqrt_and_denest_expr(self, expr: sympy.Expr) -> sympy.Expr:
        """returns denested square root of expr (with timeout)"""
        if not self.simplify or os.name == "nt":
            return sympy.sqrt(expr)
        try:
            with timeout(self.simplify_timeout, exception=SimplificationTimeOut):
                return sqrtdenest(sympy.sqrt(expr))
        except SimplificationTimeOut:
            print(f"""Simplification step timed out (>{self.simplify_timeout}s)!
Consider increasing the simpliciation time-out time or turning off simplifications.""")
            return sympy.sqrt(expr)

    def print(self, msg: str) -> None:
        """prints msg to stdout if self.silent is not True"""
        if not self.silent:
            print(msg)

    def display(self, expr: sympy.Expr, lhs: str | None = None) -> None:
        """displays sympy expression if self.silent is not True. If lhs is not None,
        the result will be formatted as
          lhs = expr
        """
        if self.silent:
            return
        eq = sympy.Eq(lhs, expr, evaluate=False) if lhs is not None else expr
        try:
            from IPython.display import display

            display(eq)
        except NameError:
            sympy.pprint(eq)

    def build(self, guesses: list[list[sympy.Expr]] | None = None) -> InflationModel:
        """Constructs an `InflationModel` instance.

        Performs fully symbolic calculation of the components of the covariant
        Hesse matrix of the potential with respect to the metric, which are then
        projected onto an orthonormal vielbein basis constructed (using the
        Gramm-Schmidt procedure) from:
          1. the gradient of the potential.
          2. the list of vectors supplied by the caller of this function.
        The details of this procedure are specified in the documentation of the
        individual methods of this class.

        ### Args
        `guesses` (`list[list[sympy.Expr]] | None`): list of vectors to be used to calculate
          an orthonormal vielbein basis onto which the components of the Hesse matrix
          will be projected. The supplied vectors *do not* have to be orthogonal, but
          they *must be* linearly independent. If this argument equals `None` and the model under
          consideration is only two-dimensional, the code is able to determine the basis on its own.
          Otherwise, it will throw an error.

        ### Returns
        `InflationModel`
        """
        if guesses is not None:
            assert len(guesses) == self.dim - 1, (
                "number of guessed vectors must equal the number of fields minus one (n-1)"
            )

        # Calculate an orthonormal basis
        # ...starting with a vector parallel to the potential gradient
        self.print("Calculating orthonormal basis...")
        self.print("Computing normalised potential gradient...")
        basis = [self.calc_v()]
        self.display(sympy.Matrix(basis[0]), lhs=sympy.symbols("v"))

        if guesses is None and self.dim == 2:
            w = [0, 0]
            w_cov = [-basis[0][1], basis[0][0]]
            metric_inv = sympy.Matrix(self.metric).inv()
            for a in range(self.dim):
                for b in range(self.dim):
                    w[a] += metric_inv[a, b] * w_cov[b]
            basis.append(self.normalize(w))
            self.display(sympy.Matrix(basis[-1]), lhs=sympy.symbols("w_1"))
        elif guesses is None:
            raise Exception("guesses argument cannot be None if model has more than two fields")
        else:
            # followed by other gramm-schmidt produced vectors
            for i, guess in enumerate(guesses):
                basis.append(self.gramm_schmidt(basis, guess))
                self.display(sympy.Matrix(basis[-1]), lhs=sympy.symbols(f"w_{i + 1}"))

        # make sure the basis is orthonormal
        if self.assertions:
            for a in range(self.dim):
                for b in range(a, self.dim):
                    try:
                        if a == b:
                            print(f"Testing if |w{a}|^2 = 1")
                            assert sympy.Eq(1, self.inner_prod(basis[a], basis[b])).simplify(), (
                                f"normalisation error: |w{a}|^2 does not equal 1"
                            )
                        else:
                            print(f"Testing if w{a}•w{b} = 0")
                            assert sympy.Eq(0, self.inner_prod(basis[a], basis[b])).simplify(), (
                                f"orthogonality error: w{a}•w{b} does not equal 0"
                            )
                    except TypeError:
                        raise Exception(
                            f"normalisation error: |w{a}|² does not equal 1"
                            if a == b
                            else f"orthogonality error: w{a}•w{b} does not equal 0"
                        )

        # Calculate the components of the covariant Hesse Matrix
        self.print("Calculating covariant Hesse matrix...")
        H = self.calc_hesse()
        self.display(sympy.Matrix(H), lhs=sympy.symbols("H"))

        # Project Hesse matrix
        def process(a: int, b: int):
            Hab = 0
            for x in range(self.dim):
                for y in range(self.dim):
                    Hab = Hab + H[x][y] * basis[a][x] * basis[b][y]
            return ([a, b], self.simplify_expr(Hab))

        self.print("Projecting the Hesse matrix on the vielbein basis...")
        H_proj = [[0 for _ in range(self.dim)] for _ in range(self.dim)]
        results = Parallel(n_jobs=cpu_count())(
            delayed(process)(a, b) for a in range(self.dim) for b in range(self.dim)
        )

        # print projected components of the Hesse matrix
        for idx, component in results:
            a, b = idx
            H_proj[a][b] = component
            if a == 0:
                a = "v"
            if b == 0:
                b = "v"
            self.display(component, lhs=sympy.symbols(f"H_{{{a}{b}}}"))

        # calculate the size of the gradient
        self.print("Calculating the norm of the gradient...")
        gradnorm = self.calc_gradient_square()
        self.display(gradnorm, lhs=Gradient(sympy.symbols("V")) ** 2)

        # compute the equations of motion
        self.print("Computing the equations of motion...")
        eoms = self.compute_eom()
        for idx, eom in enumerate(eoms):
            self.display(sympy.symbols(f"\\ddot{{{sympy.latex(self.fields[idx])}}}") + eom, lhs=0)
        constraint_h = self.compute_eom_h()
        eom_h = self.compute_eom_hdot()
        self.display(constraint_h, lhs=sympy.symbols("H"))
        self.display(eom_h, lhs=sympy.symbols("\\dot{{H}}"))

        return InflationModel(
            model_name=self.model_name,
            coordinates=self.fields,
            tangents=self.field_derivatives,
            basis=basis,
            eom_fields=eoms,
            eom_h=constraint_h,
            eom_hdot=eom_h,
            potential=self.V,
            metric=self.metric,
            gradient_square=gradnorm,
            hesse_cmp=H_proj,
        )

    def inner_prod(self, v1: list[sympy.Expr], v2: list[sympy.Expr]) -> sympy.Expr:
        """returns the inner product of v1 and v2 with respect to configured metric.

        ### Args
        - `v1` (`list[sympy.Expr]`): first vector, once contravariant
        - `v2` (`list[sympy.Expr]`): second vector, once contravariant

        ### Returns
        `sympy.Expr`: inner product of v1 with v2 with respect to the configured
          metric tensor of the current instance.
        """
        dot = 0
        for a in range(self.dim):
            for b in range(self.dim):
                dot += self.metric[a][b] * v1[a] * v2[b]
        return self.expand_and_factor_expr(dot)

    def normalize(self, vec: list[sympy.Expr]) -> list[sympy.Expr]:
        """normalizes the input vector with respect to the configured metric tensor.

        ### Args
        vec (`list[sympy.Expr]`): components of the vector to be normalised.

        ### Returns
        `list[sympy.Expr]`: normalized components of the supplied vector vec with
          respect to the metric tensor of the current instance.
        """
        # first compute the norm squared and write it as a single fraction
        normsq = 0
        for a in range(self.dim):
            for b in range(self.dim):
                normsq += self.metric[a][b] * vec[a] * vec[b]
        normsq = sympy.cancel(normsq) if self.simplify else normsq

        # Try to get rid of nested square roots
        numerator, denominator = sympy.fraction(normsq)
        numerator = self.sqrt_and_denest_expr(numerator)
        denominator = self.sqrt_and_denest_expr(denominator)

        return [
            sympy.cancel(vi * denominator / numerator)
            if self.simplify
            else vi * denominator / numerator
            for vi in vec
        ]

    def christoffels(self) -> list[list[list[sympy.Expr]]]:
        """Computes the Christoffel symbols from the metric tensor.

        Returns:
            list[list[list[sympy.Symbols]]]: Christoffel Connection Γ^a_bc, indexed as Γ[a][b][c]
        """
        # Implementation taken from Einsteinpy's christoffel.py (licensed under MIT)
        gammas = np.zeros((self.dim, self.dim, self.dim), dtype=int).tolist()
        mat = sympy.Matrix(self.metric)
        matinv = mat.inv()
        for t in range(self.dim**3):
            # i,j,k each goes from 0 to (self.dim-1)
            # hack for codeclimate. Could be done with 3 nested for loops
            k = t % self.dim
            j = (int(t / self.dim)) % (self.dim)
            i = (int(t / (self.dim**2))) % (self.dim)
            if k <= j:
                tmpvar = 0
                for n in range(self.dim):
                    tmpvar += (matinv[i, n] / 2) * (
                        sympy.diff(mat[n, j], self.fields[k])
                        + sympy.diff(mat[n, k], self.fields[j])
                        - sympy.diff(mat[j, k], self.fields[n])
                    )
            gammas[i][j][k] = gammas[i][k][j] = self.simplify_expr(tmpvar)
        return gammas

    def calc_hesse(self) -> list[list[sympy.Expr]]:
        """returns the components of the covariant Hesse matrix in a twice-covariant
        form. Components for all pairs of the supplied coordinates are calculated for
        the scalar potential V using the supplied metric tensor.

        ### Precise formulation of calculated quantities
        The components of the covariant Hesse matrix are defined as:
          V_ab(ϕ) = ∇_a ∇_b V(ϕ)
        This is expanded as (using Einstein notation):
          V_ab(ϕ) = ∇_a (∂_b V(ϕ)) = ∂_a ∂_b V(ϕ) - Γ_ab^c(ϕ) ∂_c V(ϕ)
        Where Γ_ab^c(ϕ) is the standard Christoffel connection defined as:
          Γ_ab^c = 1/2 g^cd (∂_a g_bd + ∂_b g_ad - ∂_d g_ab)

        ### Returns
        `list[list[sympy.Expr]]`: nested list of components of the covariant Hesse matrix.
        """
        # The connection has indices up-down-down (opposite order that we usually use)
        conn = self.christoffels()
        # output components of the Hesse matrix
        Vab = [[0 for _ in range(self.dim)] for _ in range(self.dim)]

        for a in range(self.dim):
            for b in range(self.dim):
                # Calculate ∂_a ∂_b V(ϕ)
                da_dbV = sympy.diff(self.V, self.fields[b], self.fields[a])

                # Calculate the contraction Γ_ab^c(ϕ) ∂_c V(ϕ)
                gamma_ab = 0
                for c in range(self.dim):
                    # Calculate ∂_c V(ϕ)
                    Vc = sympy.diff(self.V, self.fields[c])

                    # Calculate the full thing
                    gamma_ab = gamma_ab + conn[c][b][a] * Vc

                # set the output components
                cmp = da_dbV - gamma_ab
                Vab[a][b] = self.simplify_expr(cmp)
        return Vab

    def calc_gradient_square(self) -> sympy.Expr:
        """Calculates the size of the gradient of the potential given the metric g_ab.

        ### Precise formulation of calculated quantities
          output(ϕ) = g^ab(ϕ) ∂_a V(ϕ) ∂_b V(ϕ)
        Note that the output is actual the size of the gradient *squared*.

        Returns:
          sympy.Expr: size of the gradient squared (V_a V^a)
        """
        dim = len(self.fields)
        # non-normalised components of grad V
        gradient = [sympy.diff(self.V, φ) for φ in self.fields]
        out = 0.0

        # contract v with the inverse of the metric tensor
        metric_inv = sympy.Matrix(self.metric).inv()
        for a in range(dim):
            for b in range(dim):
                out += metric_inv[a, b] * gradient[a] * gradient[b]
        if os.name == "nt":
            return self.simplify_expr(out)
        try:
            with timeout(self.simplify_timeout, exception=SimplificationTimeOut):
                out = sympy.factor(sympy.expand(out))
        except SimplificationTimeOut:
            print(f"""Simplification step timed out (>{self.simplify_timeout}s)!
Consider increasing the simpliciation time-out time or turning off simplifications.""")
        return self.simplify_expr(out)

    def calc_v(self) -> list[sympy.Expr]:
        """calculates a normalized vector pointing in the direction of the gradient of
        the configured scalar potential of the current instance.

        ### Precise formulation of calculated quantities
        The contravariant components of the gradient of V are given by:
          (grad V)^a(ϕ) = g^ab(ϕ) ∂_b V(ϕ)

        ### Returns
        `list[sympy.Expr]`: contravariant components of normalized gradient vector v.
        """
        # first construct and normalize v, then raise it (this is often simpler)
        v = [sympy.diff(self.V, φ) for φ in self.fields]

        # contract v with the inverse of the metric tensor
        metric_inv = sympy.Matrix(self.metric).inv()
        vup = [0 for _ in v]
        for a in range(self.dim):
            for b in range(self.dim):
                vup[a] += metric_inv[a, b] * v[b]
        vup = self.normalize(vup)
        return [self.simplify_expr(vupi) for vupi in vup]

    def gramm_schmidt(
        self, current_basis: list[list[sympy.Expr]], guess: list[sympy.Expr]
    ) -> list[sympy.Expr]:
        """Use the Gramm-Schmidt procedure to find a new orthogonal basis vector given
        an incomplete set of orthogonal basis vectors and a third vector that is linearly
        independent from the other vectors.

        ### Args
        - `current_basis` (`list[list[sympy.Expr]]`): list of current *orthogonal*
          basisvectors. The components of the vectors should be given in
          *contravariant* form.
        - `guess` (`list[sympy.Expr]`): vector that is linearly independent from the
          (incomplete) set of current basis vectors. The components of this vector
          should be given in *contravariant* form. This vector needn't be
          normalized nor orthogonal to the set of current basis vectors.

        ### Precise formulation of calculated quantities
        The Gramm-Schmidt procedure starts with a(n incomplete) set of orthonormal
        basis vectors x_i and new vector y that is linearly independent of all x_i. We
        then simply subtract the overlap between y and each basisvector to obtain a
        vector x_i+1 that is orthogonal to all x_i:
          x_i+1 = y - Σ_a g_ij x^i_a y^j
        The final step is to normalise x_i+1

        ### Returns
        `list[sympy.Expr]`: list of the contravariant components of an additional
          basis vector orthogonal to the supplied set of basis vectors, with respect
          to the supplied metric.
        """
        dim = len(current_basis[0])
        # make sure the supplied basis is not already complete
        assert len(current_basis) < dim, "current basis is already complete. No need for more vecs."

        # start with vector y
        y = guess

        # subtract the overlap of each current basis with the guessed vector from y
        for x in current_basis:
            xdoty = self.inner_prod(x, y)
            for a in range(dim):
                y[a] -= xdoty * x[a]
        # Disable simplification if we're on Windows
        if os.name == "nt":
            return [self.simplify_expr(yi) for yi in self.normalize(y)]
        try:
            with timeout(self.simplify_timeout, exception=SimplificationTimeOut):
                y = [sympy.factor(sympy.expand(yi)) for yi in y]
        except SimplificationTimeOut:
            print(f"""Simplification step timed out (>{self.simplify_timeout}s)!
Consider increasing the simpliciation time-out time or turning off simplifications.""")

        return [self.simplify_expr(yi) for yi in self.normalize(y)]

    def project_hesse(
        self,
        hesse_matrix: list[list[sympy.Expr]],
        v1: list[sympy.Expr],
        v2: list[sympy.Expr],
    ) -> sympy.Expr:
        """This function calculates the projection of the Hesse matrix along the two
        supplied vectors.

        ### Args
        - `hesse_matrix` (`list[list[sympy.Expr]]`): twice-covariant components of
          the Hesse matrix.
        - `v1` (`list[sympy.Expr]`): first vector along which to project the Hesse
          matrix.
        - `v2` (`list[sympy.Expr]`): second vector along which to project the Hesse
          matrix.

        ### Precise formulation of calculated quantities
        The output H12 is given by
          H12 = H_ab v1^a v2^b
        Where v1 is the first input vector, v2 the second one and H_ab are the
        twice covariant components of the Hesse matrix. No metrics are required for
        this operation.

        ### Returns
        `sympy.Expr`: returns the inner product of the Hesse matrix with v1 and v2.
        """
        V_proj = 0
        for a in range(self.dim):
            for b in range(self.dim):
                V_proj = V_proj + hesse_matrix[a][b] * v1[a] * v2[b]
        return self.simplify_expr(V_proj)

    def compute_eom(self):
        """This function calculates the equations of motion for the second-order time derivatives
        of the coordinate fields.

        ### Precise formulation of calculated quantities
        The multifield equations of motion can be recast in the following form:
             χ'^a + Γ^a_bc χ^b χ^c + ∇V = 0
             φ'^a - χ^a = 0
             H' + 3H^2 - V = 0
        The components of the first equation are computed by this function for all fields.
        """
        connection = self.christoffels()
        out = [0 for _ in range(self.dim)]
        metric_inv = sympy.Matrix(self.metric).inv()
        for a in range(self.dim):
            conn_term = 0
            grad_term = 0
            for b in range(self.dim):
                for c in range(self.dim):
                    conn_term += (
                        connection[a][b][c] * self.field_derivatives[b] * self.field_derivatives[c]
                    )
                grad_term += metric_inv[a, b] * sympy.diff(self.V, self.fields[b])
            conn_term = self.expand_and_factor_expr(conn_term)
            grad_term = self.expand_and_factor_expr(grad_term)
            out[a] = conn_term + grad_term
        return [self.simplify_expr(outi) for outi in out]

    def compute_eom_h(self):
        """This function calculates the constraint equation for the Hubble parameter.

        ### Precise formulation of calculated quantities
        We compute H from the fields and their derivatives:
            3H² = V + 1/2 G_ab χ^a χ^b
            => H = √[(V + 1/2 G_ab χ^a χ^b) / 3]
        """
        out = self.V
        for a in range(self.dim):
            for b in range(self.dim):
                out += self.metric[a][b] * self.field_derivatives[a] * self.field_derivatives[b]
        out = self.expand_and_factor_expr(out)
        return self.sqrt_and_denest_expr(out / 3)

    def compute_eom_hdot(self):
        """This function computes the equation of motion for the Hubble parameter.

        ### Precise formulation of calculated quantities
        We compute H' from the fields and their derivatives:
            H' = -1/2 G_ab χ^a χ^b
        """
        out = 0
        for a in range(self.dim):
            for b in range(self.dim):
                out -= self.metric[a][b] * self.field_derivatives[a] * self.field_derivatives[b]
        out /= sympy.nsimplify(2)
        return self.expand_and_factor_expr(out)



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# ************************************************************************************************ #
# Angular Inflation used in Wolters, Iarygina and Achúcarro JCAP07(2024)079
# Model from Kenton and Thomas JCAP02(2015)127 and Chakraborty, Chiovoloni et al. JCAP01(2020)020
# ************************************************************************************************ #

import os

import inflatox
import numpy as np
import sympy
from inflatox.consistency_conditions import GeneralisedAL
from sympy.simplify.radsimp import collect_sqrt

trajectory_dir = f"{os.path.dirname(os.path.abspath(__file__))}/trajectories/"


def test_egno():
    model_name = "d5"

    # setup model
    r, θ = sympy.symbols("r θ2")
    fields = [r, θ]

    gs, ls, N = sympy.symbols("g_s l_s N")
    mu5, T5, Lt = sympy.symbols("mu5 T5 L_T")

    mu5 = 1 / ((2 * sympy.pi) ** 5 * ls**6)
    T5 = mu5 / gs

    rho, u = sympy.symbols("rho u")
    rho = r / (3 * u)

    H = (
        ((sympy.pi * N * gs * ls**4) / (12 * u**4) * (2 / rho**2 - 2 * sympy.ln(1 / rho**2 + 1)))
        .nsimplify()
        .collect([u, r])
        .expand()
        .powsimp(force=True)
    )

    p, q = sympy.symbols("p q")

    F = (
        (H / 9 * (r**2 + 3 * u**2) ** 2 + (sympy.pi * q * ls**2) ** 2)
        .nsimplify()
        .collect([r, u])
        .expand()
        .powsimp()
    )

    gamma = 4 * sympy.pi**2 * ls**2 * p * q * T5 * gs

    sqrtF = sympy.sqrt(F)
    g00 = (
        collect_sqrt(
            4 * sympy.pi * p * T5 * sqrtF * ((r**2 + 6 * u**2) / (r**2 + p * u**2)),
            evaluate=True,
        )
        .expand()
        .powsimp()
    )
    g11 = (
        collect_sqrt((4 / 6) * sympy.pi * p * T5 * sqrtF * (r**2 + 6 * u**2), evaluate=True)
        .nsimplify()
        .collect([r, u])
        .expand()
        .powsimp()
    )

    metric = [[g00, 0], [0, g11]]

    Phi_min = (
        (
            (5 / 72)
            * (
                81 * (9 * rho**2 - 2) * rho**2
                + 162 * sympy.ln(9 * (rho**2 + 1))
                + -9
                + -160 * sympy.ln(10)
            )
        )
        .nsimplify()
        .collect([u])
        .expand()
        .powsimp()
    )

    a0, a1, b1 = sympy.symbols("a0 a1 b1")
    Phi_h = (
        (
            a0 * (2 / rho**2 - 2 * sympy.ln(1 / rho**2 + 1))
            + 2
            * a1
            * (6 + 1 / rho**2 - 2 * (2 + 3 * rho**2) * sympy.ln(1 + 1 / rho**2))
            * sympy.cos(θ)
            + (b1 / 2) * (2 + 3 * rho**2) * sympy.cos(θ)
        )
        .nsimplify()
        .collect([u, r])
        .expand()
        .powsimp()
    )

    V0 = sympy.symbols("V0")
    potential = (
        V0
        + (4 * sympy.pi * p * T5 / H) * (sympy.sqrt(F) - (ls**2) * sympy.pi * q * gs)
        + gamma * (Phi_min + Phi_h)
    )
    potential = potential.nsimplify().collect([ls, gs]).expand().powsimp()

    model = inflatox.InflationModelBuilder.new(
        fields,
        metric,
        potential,
        model_name=model_name,
        assertions=False,
        silent=True,
        simplify=False,
    ).build([[1, 0]])

    out = inflatox.Compiler(model, cleanup=False).compile()
    anguelova = GeneralisedAL(out)

    V0 = -1.17e-8
    N = 1000.0
    gs = 0.01
    ls = 501.961
    u = 50 * ls
    q = 1.0
    p = 5.0
    a0 = 0.001
    a1 = 0.0005
    b1 = 0.001
    parameters = np.array([V0, a0, p, q, u, ls, a1, b1, gs, N])

    r_start, r_stop = 0.0, 36.0
    θ_start, θ_stop = 0.0, 4 * np.pi
    extent = (r_start, r_stop, θ_start, θ_stop)
    N = 120

    # Calculate potential
    anguelova.calc_V_array(parameters, [r_start, r_stop], [θ_start, θ_stop], [N, N])

    # run analysis
    anguelova.complete_analysis(parameters, *extent, *[N, N])

    # run analysis on trajectory
    trajectory = np.loadtxt(f"{trajectory_dir}/d5_trajectory.dat")

    anguelova.complete_analysis_ot(parameters, trajectory)

    # run Anguelova's original condition
    anguelova.consistency_rapidturn(parameters, *extent, *[N, N])



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# ************************************************************************************************ #
# Angular Inflation used in Wolters, Iarygina and Achúcarro JCAP07(2024)079
# Model from Ellis, Garcia, Nanopoulos and Olive JCAP08(2014)044
# ************************************************************************************************ #

import os

import inflatox
import numpy as np
import sympy
from inflatox.consistency_conditions import GeneralisedAL

trajectory_dir = f"{os.path.dirname(os.path.abspath(__file__))}/trajectories/"


def test_egno():
    model = "egno"

    # setup model
    alpha, m, p, c, a = sympy.symbols("alpha m p c a")
    r, θ = sympy.symbols("r θ")
    fields = [r, θ]

    Phi, Phi_Bar, S, S_Bar = sympy.symbols("Phi Phi_B S S_B")
    K = (
        -3 * alpha * sympy.ln(Phi + Phi_Bar - c * (Phi + Phi_Bar - 1) ** 4)
        + (S * S_Bar) / (Phi + Phi_Bar) ** 3
    ).nsimplify()

    superfields = [Phi, S]
    superfields_conjugate = [Phi_Bar, S_Bar]
    metric = [
        [sympy.diff(sympy.diff(K, superfields[b]), superfields_conjugate[a]) for a in range(0, 2)]
        for b in range(0, 2)
    ]
    metric = [
        [g.subs({Phi: r + 1j * θ, Phi_Bar: r - 1j * θ}).nsimplify().simplify() for g in gb]
        for gb in metric
    ]
    metric = [[g.subs({S: 0, S_Bar: 0}).simplify() for g in gb] for gb in metric]
    real_metric = [[metric[0][0], 0], [0, metric[0][0]]]

    potential = (
        (6 * m**2 * r**3 * ((a - r) ** 2 + θ**2))
        / (a**2 * (2 * r - c * (1 - 2 * r) ** 4) ** (3 * alpha))
    ).nsimplify()

    model = inflatox.InflationModelBuilder.new(
        fields,
        real_metric,
        potential,
        model_name=model,
        silent=True,
        simplify=False,
        assertions=False,
    ).build([[0, 1]])

    out = inflatox.Compiler(model, silent=False, cse=True).compile()
    anguelova = GeneralisedAL(out)

    alpha = 1.0
    a = 0.5
    c = 1000.0
    p = 3.055
    m = 1e-3
    args = np.array([m, a, c, alpha])

    r_start, r_stop = 0.45, 0.55
    θ_start, θ_stop = 0.0, np.pi
    N_r, N_θ = 500, 100
    extent = (0.46, 0.50, θ_start, θ_stop)

    # Calculate potential
    anguelova.calc_V_array(args, [r_start, r_stop], [θ_start, θ_stop], [N_r, N_θ])

    # run analysis
    anguelova.complete_analysis(args, *extent, *[N_r, N_θ])

    # run analysis on trajectory
    tr = np.load(f"{trajectory_dir}/egno_r.npy")
    ttheta = np.load(f"{trajectory_dir}/egno_theta.npy")
    trajectory = np.column_stack((tr, ttheta))
    anguelova.complete_analysis_ot(args, trajectory)

    # run Anguelova's original condition
    anguelova.consistency_rapidturn(args, *extent, *[N_r, N_θ])



#  Copyright© 2023 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# External imports
import numpy as np

# Internal imports
from .compiler import CompilationArtifact
from .libinflx_rs import *

# Limit exports to these items
__all__ = ["InflationCondition", "GeneralisedAL"]


class InflationCondition:
    """Base class for all inflation conditions. Provides native methods to evaluate
    the potential and projected Hesse matrix. This base-class may be extended either
    by using these native methods, or by including your own native code that hooks
    into the Rust API or C ABI.
    """

    def __init__(self, compiled_artifact: CompilationArtifact, validate_basis: bool = True):
        """Passes compiled model to lib_inflx_rs.

        ### Args:
        - `compiled_artifact` (`CompilationArtifact`): output of `Compiler` (see its docs)
        - `validate_basis` (`bool`, optional): if `True`, lib_inflx_rs will check that the
          field-space basis defined in the `CompilationArtifact` is orthonormal
          at some number of random field-space points for with random parameter values. It will
          throw an exception if this is not the case. You may disable this if inflatox picks
          random points outside the domain of your model.
        """
        self.artifact = compiled_artifact
        self.dylib = open_inflx_dylib(compiled_artifact.shared_object_path, validate_basis)

    def calc_V(self, x: np.ndarray, args: np.ndarray) -> float:
        """calculates the scalar potential at field-space coordinates `x` with
        model-specific parameters `args`.

        ### Args
        - `x` (`np.ndarray`): field-space coordinates at which to calculate
        - `args` (`np.ndarray`): values of the model-dependent parameters. See
        `CompilationArtifact.print_sym_lookup_table()` for an overview of which
        sympy symbols were mapped to which args index.

        ### Returns
          `float`: Value of scalar potential with parameters `args` at coordinates `x`.
        """
        return self.dylib.potential(x, args)

    def calc_V_array(
        self,
        args: list[float] | np.ndarray,
        start: list[float] | np.ndarray,
        stop: list[float] | np.ndarray,
        N: list[int] | None = None,
    ) -> np.ndarray:
        """constructs an array of field space coordinates and fills it with the
        value of the scalar potential at those field space coordinates.
        The start and stop values of each axis in field-space can be specified with
        the `start` and `stop` arguments. The number of samples along each axis can
        be set with the `N` argument. It defaults to `8000` per axis.

        ### Args
        - `args` (`list[float] | np.ndarray`): values of the model-dependent
        parameters. See `CompilationArtifact.print_sym_lookup_table()` for an
        overview of which sympy symbols were mapped to which args index.
        - `start` (`list[float] | np.ndarray`): list of minimum values for
        each axis of the to-be-constructed array in field space.
        - `stop` (`list[float] | np.ndarray`): list of maximum values for each
        axis of the to-be-constructed array in field space.
        - `N` (`list[int] | None`, optional): _description_. list of the number of
        samples along each axis in field space. If set to `None`, 8000 samples will
        be used along each axis.

        ### Returns
        `np.ndarray`: value of scalar potential at specified field-space
        coordinates
        """
        n_fields = self.artifact.n_fields
        start_stop = np.array([[float(start), float(stop)] for (start, stop) in zip(start, stop)])
        N = N if N is not None else [8000 for _ in range(n_fields)]
        x = np.zeros(N)
        self.dylib.potential_array(x, args, start_stop)
        return x

    def calc_H(self, x: np.ndarray, args: np.ndarray) -> np.ndarray:
        """calculates the projected covariant Hesse matrix at field-space
        coordinates `x` with model-specific parameters `args`.

        ### Args
        - `x` (`np.ndarray`): field-space coordinates at which to calculate
        - `args` (`np.ndarray`): values of the model-dependent parameters. See
        `CompilationArtifact.print_sym_lookup_table()` for an overview of which
        sympy symbols were mapped to which args index.

        ### Returns
        `np.ndarray`: Components of the projected covariant hesse matrix with
          parameters `args` at coordinates `x`.
        """
        return self.dylib.hesse(x, args)

    def calc_H_array(
        self,
        args: list[float] | np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N: list[int] | None = None,
    ) -> np.ndarray:
        """constructs an array of field space coordinates and fills it with the
        value of the projected Hesse matrix at those field space coordinates.
        The start and stop values of each axis in field-space can be specified with
        the `start` and `stop` arguments. The number of samples along each axis can
        be set with the `N` argument. It defaults to `8000` per axis.

        ### Args
        - `args` (`list[float] | np.ndarray`): values of the model-dependent
        parameters. See `CompilationArtifact.print_sym_lookup_table()` for an
        overview of which sympy symbols were mapped to which args index.
        - `start` (`list[float] | np.ndarray`): list of minimum values for
        each axis of the to-be-constructed array in field space.
        - `stop` (`list[float] | np.ndarray`): list of maximum values for each
        axis of the to-be-constructed array in field space.
        - `N` (`list[int] | None`, optional): _description_. list of the number of
        samples along each axis in field space. If set to `None`, 8000 samples will
        be used along each axis.

        ### Returns
        `np.ndarray`: (d+2)-dimensional array for a d-dimensional field-space. The
          first two axes of this array represent the axes of the Hesse matrix itself.
          The other axes correspond to the field-space components.
        """
        n_fields = self.artifact.n_fields
        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )
        N = N if N is not None else [8000 for _ in range(n_fields)]
        return self.dylib.hesse_array(np.array(N, dtype=np.uintp), args, start_stop)

    def validate_basis_on_domain(
        self,
        args: list[float] | np.ndarray,
        start: list[float] | np.ndarray,
        stop: list[float] | np.ndarray,
        N: list[int] | int = 100,
        accuracy: float = 1e-3,
    ) -> None:
        """checks if basis is orthonormal at all points in the interval [start, stop].

        If `start` and `stop` are arrays/lists, they will be interpreted as start/stop values along
        each axis. The number of samples along each axis is specified with the `N` argument.

        At each point, the basis defined in the model is constructed and it is verified that:
            1. All vectors are normalised in the sense that (g_ab v^a v^a - 1) < accuracy.
            2. All vectors are orthogonal in the sense that |g_ab v^a v^b| < accuracy.

        ### Args
        - `args` (`list[float] | np.ndarray`): values of the model-dependent
        parameters. See `CompilationArtifact.print_sym_lookup_table()` for an
        overview of which sympy symbols were mapped to which args index.
        - `start` (`list[float] | np.ndarray`): list of minimum values for
        each axis of the to-be-constructed array in field space.
        - `stop` (`list[float] | np.ndarray`): list of maximum values for each
        axis of the to-be-constructed array in field space.
        - `N` (`int | list[int]`, optional): list of the number of samples along each axis in field
        space. If set to a single `int`, the same number of samples will be used along each axis.
        Defaults to 100.
        - `accuracy` (float, optional): tolerance to within basis must be orthonormal. Defaults to
        1e-3.

        ### Returns
        `None`. May throw an exception if the basis is malformed.
        """
        n_fields = self.artifact.n_fields
        start_stop = np.array([[float(start), float(stop)] for (start, stop) in zip(start, stop)])
        if N is int:
            N = N * np.ones(n_fields)
        self.dylib.validate_basis_on_domain(N, args, start_stop, accuracy)


class GeneralisedAL(InflationCondition):
    """This class extends the generic `InflationCondition` with the generalised rapid-turn (ω>>ε^½)
    consistency condition from (`arXiv:2405.11628`).

    In addition, related quantities that may be estimated from the potential and field-space metric,
    such as the slow-roll parameters εH, η_||, the angle δ and turn-rate ω can also be computed.
    The rapid-turn limit of the consistency condition derived in the 2022 Anguelova and Lazaroiu 2022
    (`arXiv:2210.00031v2`) is also available.

    ### Usage
    To construct an instance of this class, a `CompilationArtifact` is required.
    Such an artifact can be obtained by running an instance of `inflatox.Compiler`
    with a specific model (fieldspace metric + scalar potential). For more info on
    how to use the `Compiler`, see its documentation.

    The artifact contains all the necessary information to evaluate the consistency
    condition and can be used to construct an instance of this class.

    To run evaluate the consistency condition for various model parameters and
    regions of field-space, use the `.complete_analysis()` method on an instance
    of this class with the appropriate methods. For more info, see the
    `.complete_analysis()` method.
    """

    def __init__(self, compiled_artifact: CompilationArtifact):
        super().__init__(compiled_artifact)

    def complete_analysis(
        self,
        args: np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N_x0: int = 1_000,
        N_x1: int = 1_000,
        progress: bool = True,
        threads: None | int = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """This function performs a complete analysis of possible slow-roll (rapid)
        turn trajectories using the methods described in (paper), based on the AL
        consistency condition. It returns six arrays filled with:
          1. Let rhs and lhs denote the left-hand side and right-hand side of the AL
            consistency condition. This first array is populated with the normalised
            difference between the rhs and lhs:
              out = ||lhs| - |rhs||/(|lhs| + |rhs|)
            Note that out = 0 corresponds to the consistency condition holding
            perfectly. Values larger than zero indicate it does not hold perfectly.
          2. ε_V (first potential slow-roll parameter)
          3. ε_H (first dynamical slow-roll parameter), calculated assuming that the
            AL condition holds.
          4. η_|| (second dynamical slow-roll parameter), calculated assuming that the
            AL condition holds.
          5. δ (characteristic angle), calculated assuming that the AL condition holds.
          6. ω (relative turn rate), calculated assuming that the AL condition holds.
        Using (1) and (3), slow-roll trajectories with |ε_H|, |η_||| << 1 can be
        identified. See (paper) for a more complete discussion.

        ### *NOTE*
        The consistency condition or ε_H *on their own* are usually insufficient to discriminate
        determine the location of the rapid-turn attractor. Always verify that:
        1. The consistency condition holds AND
        2. ε_H << 1 AND
        3. ε_H > 0

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x0_start` (`float`): minimum value of first field `x[0]`.
        - `x0_stop` (`float`): maximum value of first field `x[0]`.
        - `x1_start` (`float`): minimum value of second field `x[1]`.
        - `y_stop` (`float`): maximum value of second field `x[1]`.
        - `N_x` (`int`, optional): number of steps along `x[0]` axis. Defaults to 10_000.
        - `x1_stop` (`int`, optional): number of steps along `x[1]` axis. Defaults to 10_000.
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray` (✕6): arrays filled with the following quantities (in order):
          1. Consistency condition ||lhs| - |rhs||/(|lhs| + |rhs|)
          2. ε_V (first potential slow-roll parameter)
          3. ε_H (first dynamical slow-roll parameter)
          4. η_|| (speed-up parameter)
          5. δ (characteristic angle)
          6. ω (relative turn rate)
        """

        # set up args for anguelova's condition
        out = np.zeros((N_x0, N_x1, 6), dtype=float)

        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )

        # calculate
        threads = threads if threads is not None else 0

        # evaluate and return
        complete_analysis(self.dylib, args, out, start_stop, progress, threads)
        return (
            out[:, :, 0],
            out[:, :, 1],
            out[:, :, 2],
            out[:, :, 3],
            out[:, :, 4],
            out[:, :, 5],
        )

    def consistency(
        self,
        args: np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N_x0: int = 1_000,
        N_x1: int = 1_000,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the normalised difference between one and the
        quotient of the left-hand-side (lhs) and right-hand-side (rhs) of the slow-
        roll turn consistency condition.

        ### Exact formulation of calculated quantities
        This function returns:
          ||lhs| - |rhs||/(|lhs| + |rhs|)
        Where
          lhs = Vww/V
          rhs = 3 + 3 (Vvw/Vvv)² + (Vvv/V) (Vvw/Vvv)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x0_start` (`float`): minimum value of first field `x[0]`.
        - `x0_stop` (`float`): maximum value of first field `x[0]`.
        - `x1_start` (`float`): minimum value of second field `x[1]`.
        - `y_stop` (`float`): maximum value of second field `x[1]`.
        - `N_x` (`int`, optional): number of steps along `x[0]` axis. Defaults to 10_000.
        - `x1_stop` (`int`, optional): number of steps along `x[1]` axis. Defaults to 10_000.
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with slow-roll (intermediate) turn consistency
          condition from (paper)
        """

        # set up args for anguelova's condition
        out = np.zeros((N_x0, N_x1), dtype=float)

        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )

        # calculate
        threads = threads if threads is not None else 0

        # evaluate and return
        consistency_only(self.dylib, args, out, start_stop, progress, threads)
        return out

    def epsilon_v(
        self,
        args: np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N_x0: int = 1_000,
        N_x1: int = 1_000,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the potential first-order slow-roll parameter
        ε_V

        ### Exact formulation of calculated quantities
        This function returns:
          ε_V = 1/2 (∇V/V)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x0_start` (`float`): minimum value of first field `x[0]`.
        - `x0_stop` (`float`): maximum value of first field `x[0]`.
        - `x1_start` (`float`): minimum value of second field `x[1]`.
        - `y_stop` (`float`): maximum value of second field `x[1]`.
        - `N_x` (`int`, optional): number of steps along `x[0]` axis. Defaults to 10_000.
        - `x1_stop` (`int`, optional): number of steps along `x[1]` axis. Defaults to 10_000.
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with slow-roll (intermediate) turn consistency
          condition from (paper)
        """

        # set up args for anguelova's condition
        out = np.zeros((N_x0, N_x1), dtype=float)

        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )

        # calculate
        threads = threads if threads is not None else 0

        # evaluate and return
        epsilon_v_only(self.dylib, args, out, start_stop, progress, threads)
        return out

    def consistency_rapidturn(
        self,
        args: np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N_x0: int = 1_000,
        N_x1: int = 1_000,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the normalised difference between one and the
        quotient of the left-hand-side (lhs) and right-hand-side (rhs) of
        Anguelova & Lazaroiu's original consistency condition (`arXiv:2210.00031v2`).

        ### Exact formulation of calculated quantities
        This function returns:
          ||lhs| - |rhs||/(|lhs| + |rhs|)
        Where
          lhs = Vww/V
          rhs = 3 (Vvw/Vvv)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x0_start` (`float`): minimum value of first field `x[0]`.
        - `x0_stop` (`float`): maximum value of first field `x[0]`.
        - `x1_start` (`float`): minimum value of second field `x[1]`.
        - `y_stop` (`float`): maximum value of second field `x[1]`.
        - `N_x` (`int`, optional): number of steps along `x[0]` axis. Defaults to 10_000.
        - `x1_stop` (`int`, optional): number of steps along `x[1]` axis. Defaults to 10_000.
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with Anguelova & Lazaroiu's original consistency
          condition.
        """

        # set up args for anguelova's condition
        out = np.zeros((N_x0, N_x1), dtype=float)

        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )

        # calculate
        threads = threads if threads is not None else 0

        # evaluate and return
        consistency_rapidturn_only(self.dylib, args, out, start_stop, progress, threads)
        return out

    def flag_quantum_dif(
        self,
        args: np.ndarray,
        x0_start: float,
        x0_stop: float,
        x1_start: float,
        x1_stop: float,
        N_x0: int = 10_000,
        N_x1: int = 10_000,
        progress=True,
        accuracy=1e-3,
    ) -> np.ndarray:
        """returns boolean array where `True` values indicate that both components
        of the gradient of the scalar potential are smaller than the specified
        `accuracy` parameter. This is useful to identify points in the potential where
        quantum diffusion may have a large impact (saddle points in the potential).
        This calculation explicitly *does not* take into account the full inner product
        using the metric to avoid measuring where the metric goes to zero or becomes
        signular.

        Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x0_start` (`float`): minimum value of first field `x[0]`.
        - `x0_stop` (`float`): maximum value of first field `x[0]`.
        - `x1_start` (`float`): minimum value of second field `x[1]`.
        - `y_stop` (`float`): maximum value of second field `x[1]`.
        - `N_x` (`int`, optional): number of steps along `x[0]` axis. Defaults to 10_000.
        - `x1_stop` (`int`, optional): number of steps along `x[1]` axis. Defaults to 10_000.
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `accuracy` (`float`, optional):

        Returns:
        `np.ndarray`: boolean array. `True` where the absolute value of both components
          of the gradient are smaller than `accuracy`, `False` otherwise.
        """

        # set up args for anguelova's condition
        x = np.zeros((N_x0, N_x1), dtype=bool)

        start_stop = np.array(
            [[float(x0_start), float(x0_stop)], [float(x1_start), float(x1_stop)]]
        )

        # evaluate and return
        flag_quantum_dif_py(self.dylib, args, x, start_stop, progress, accuracy)
        return x

    #########################
    # On_trajectory methods #
    #########################

    def complete_analysis_ot(
        self,
        args: np.ndarray,
        x: np.ndarray,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """This function performs a complete analysis of possible slow-roll (rapid)
        turn trajectories using the methods described in (paper), based on the AL
        consistency condition. It returns six arrays filled with:
          1. Let rhs and lhs denote the left-hand side and right-hand side of the AL
            consistency condition. This first array is populated with the normalised
            difference between the rhs and lhs:
              out = ||lhs| - |rhs||/(|lhs| + |rhs|)
            Note that out = 0 corresponds to the consistency condition holding
            perfectly. Values larger than zero indicate it does not hold perfectly.
          2. ε_V (first potential slow-roll parameter)
          3. ε_H (first dynamical slow-roll parameter), calculated assuming that the
            AL condition holds.
          4. η_|| (second dynamical slow-roll parameter), calculated assuming that the
            AL condition holds.
          5. δ (characteristic angle), calculated assuming that the AL condition holds.
          6. ω (relative turn rate), calculated assuming that the AL condition holds.
        Using (1) and (3), slow-roll trajectories with |ε_H|, |η_||| << 1 can be
        identified. See (paper) for a more complete discussion.

        ### *NOTE*
        The consistency condition or ε_H *on their own* are usually insufficient to discriminate
        determine the location of the rapid-turn attractor. Always verify that:
        1. The consistency condition holds AND
        2. ε_H << 1 AND
        3. ε_H > 0

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x` (`np.ndarray`): array of field-space points specifying the trajectory
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually 1).
          When set to 1, a single-threaded implementation will be used.

        ### Returns:
        `np.ndarray` (✕6): arrays filled with the following quantities (in order):
          1. Consistency condition ||lhs| - |rhs||/(|lhs| + |rhs|)
          2. ε_V (first potential slow-roll parameter)
          3. ε_H (first dynamical slow-roll parameter)
          4. η_|| (speed-up parameter)
          5. δ (characteristic angle)
          6. ω (relative turn rate)
        """

        # set up args for anguelova's condition
        out = np.zeros((x.shape[0], 6), dtype=float)

        # Single-threaded default is more appropriate for smaller number of iterations
        threads = threads if threads is not None else 1

        # evaluate and return
        complete_analysis_on_trajectory(self.dylib, args, x, out, progress, threads)
        return np.split(out, 6, 1)

    def consistency_ot(
        self,
        args: np.ndarray,
        x: np.ndarray,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the normalised difference between one and the
        quotient of the left-hand-side (lhs) and right-hand-side (rhs) of the slow-
        roll turn consistency condition.

        ### Exact formulation of calculated quantities
        This function returns:
          ||lhs| - |rhs||/(|lhs| + |rhs|)
        Where
          lhs = Vww/V
          rhs = 3 + 3 (Vvw/Vvv)² + (Vvv/V) (Vvw/Vvv)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x` (`np.ndarray`): array of field-space points specifying the trajectory
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with slow-roll (intermediate) turn consistency
          condition from (paper)
        """
        # set up args for anguelova's condition
        out = np.zeros((x.shape[0]), dtype=float)

        # Single-threaded default is more appropriate for smaller number of iterations
        threads = threads if threads is not None else 1

        # evaluate and return
        consistency_only_on_trajectory(self.dylib, args, x, out, progress, threads)
        return out

    def consistency_rapidturn_ot(
        self,
        args: np.ndarray,
        x: np.ndarray,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the normalised difference between one and the
        quotient of the left-hand-side (lhs) and right-hand-side (rhs) of the slow-
        roll turn consistency condition.

        ### Exact formulation of calculated quantities
        This function returns:
          ||lhs| - |rhs||/(|lhs| + |rhs|)
        Where
          lhs = Vww/V
          rhs = 3 + 3 (Vvw/Vvv)² + (Vvv/V) (Vvw/Vvv)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x` (`np.ndarray`): array of field-space points specifying the trajectory
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with slow-roll (intermediate) turn consistency
          condition from (paper)
        """
        # set up args for anguelova's condition
        out = np.zeros((x.shape[0]), dtype=float)

        # Single-threaded default is more appropriate for smaller number of iterations
        threads = threads if threads is not None else 1

        # evaluate and return
        consistency_rapidturn_only_on_trajectory(self.dylib, args, x, out, progress, threads)
        return out

    def epsilon_v_ot(
        self,
        args: np.ndarray,
        x: np.ndarray,
        progress: bool = True,
        threads: None | int = None,
    ) -> np.ndarray:
        """returns array filled with the normalised difference between one and the
        quotient of the left-hand-side (lhs) and right-hand-side (rhs) of
        Anguelova & Lazaroiu's original consistency condition (`arXiv:2210.00031v2`).

        ### Exact formulation of calculated quantities
        This function returns:
          ||lhs| - |rhs||/(|lhs| + |rhs|)
        Where
          lhs = Vww/V
          rhs = 3 (Vvw/Vvv)²

        ### Args:
        - `args` (`np.ndarray`): values of the model-dependent parameters.
        - `x` (`np.ndarray`): array of field-space points specifying the trajectory
        - `progress` (`bool`, optional): whether to render a progressbar or not. Showing the
          progressbar may slightly degrade performance. Defaults to True.
        - `threads` (`None | int`, optional): number of threads to use for calculation.
          When set to `None`, inflatox will choose the optimum number (usually equal
          to the number of CPU's). When set to 1, a single-threaded implementation
          will be used.

        ### Returns:
        `np.ndarray`: array filled with Anguelova & Lazaroiu's original consistency
          condition.
        """
        # set up args for anguelova's condition
        out = np.zeros((x.shape[0]), dtype=float)

        # Single-threaded default is more appropriate for smaller number of iterations
        threads = threads if threads is not None else 1

        # evaluate and return
        epsilon_v_only_on_trajectory(self.dylib, args, x, out, progress, threads)
        return out



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# External imports
import numpy as np

# Internal imports
from .compiler import CompilationArtifact
from .libinflx_rs import open_inflx_dylib, solve_eom_rk4, solve_eom_rkf

__all__ = ["solve_eom"]


def solve_eom(
    artifact: CompilationArtifact,
    pars: np.ndarray[float],
    steps: int,
    fields_init: list[float],
    derivatives_init: list[float],
    max_err: float = 1e-6,
    solver: str = "rk4",
):
    n = artifact.n_fields
    out = np.zeros((steps, n * 2 + 1))
    out[0, 0:n] = np.array(fields_init)
    out[0, n : 2 * n] = np.array(derivatives_init)

    dylib = open_inflx_dylib(artifact.shared_object_path)
    if solver == "rk4":
        solve_eom_rk4(dylib, pars, out, max_err)
    else:
        solve_eom_rkf(dylib, pars, out, max_err)
    return out



#  Copyright© 2023 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

from .compiler import CompilationArtifact, Compiler

from .symbolic import InflationModel, InflationModelBuilder

from .version import __version__

from . import consistency_conditions
from . import background
from .libinflx_rs import log_info, log_warn

__all__ = [
    "CompilationArtifact",
    "Compiler",
    "InflationModel",
    "InflationModelBuilder",
    "consistency_conditions",
    "log_info",
    "log_warn",
    "background",
    "__version__",
]



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

import inflatox
import numpy as np
import sympy as sp


def test_doc_example():
    # define model
    r, θ, m = sp.symbols("r θ m")
    fields = [r, θ]

    V = (1 / 2 * m**2 * (θ**2 - 2 / (3 * r**2))).nsimplify()
    g = [[0.5, 0], [0, 0.5 * r**2]]

    # symbolic calculation
    model = inflatox.InflationModelBuilder.new(fields, g, V).build()

    # run the compiler
    out = inflatox.Compiler(model).compile()
    out.print_sym_lookup_table()

    # evaluate the compiled potential and Hesse matrix
    from inflatox.consistency_conditions import GeneralisedAL

    anguelova = GeneralisedAL(out)

    params = np.array([1.0])
    x = np.array([2.0, -2.0])
    v_val = anguelova.calc_V(x, params)
    h_val = anguelova.calc_H(x, params)
    print(v_val, h_val)
    assert v_val == 1.9166666666666667
    assert np.allclose(h_val, np.array([[0.41206897, -1.05517241], [-1.05517241, -0.07873563]]))

    extent = [0.0, 2.5, 0.0, np.pi]
    consistency_condition, epsilon_V, epsilon_H, eta_H, delta, omega = anguelova.complete_analysis(
        params, *extent
    )

    assert np.nanmax(consistency_condition) <= 1



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

import pytest
import sympy
from inflatox import InflationModelBuilder


@pytest.fixture
def angular_model():
    f1, f2 = sympy.symbols("phi_1 phi_2")
    m1, m2, alpha = sympy.symbols("m_1 m_2 alpha")
    v = (alpha / 2) * ((m1 * f1) ** 2 + (m2 * f2) ** 2)
    diag = 6 * alpha / ((1 - f1**2 - f2**2) ** 2)
    metric = [[diag, 0], [0, diag]]
    return InflationModelBuilder.new([f1, f2], metric, v, "[test] angular inflation model")


@pytest.fixture
def trivial_model():
    f1, f2 = sympy.symbols("phi_1 phi_2")
    m1, m2 = sympy.symbols("m_1 m_2")
    v = (m1 * f1) ** 2 + (m2 * f2) ** 2
    metric = [[1, 0], [0, 1]]
    return InflationModelBuilder.new([f1, f2], metric, v, "[test] trivial inflation model")


def test_inner_prod(trivial_model):
    v1 = [1, 0]
    v2 = [0, 1]
    assert sympy.Eq(trivial_model.inner_prod(v1, v2), 0)


def test_normalize(trivial_model):
    a = sympy.symbols("a")
    v = [1, a**2]
    vnorm = trivial_model.normalize(v)
    assert sympy.Eq(trivial_model.inner_prod(vnorm, vnorm), 1).simplify()


def test_trivial_christoffels(trivial_model):
    Gamma = trivial_model.christoffels()
    for a in range(trivial_model.dim):
        for b in range(trivial_model.dim):
            for c in range(trivial_model.dim):
                assert sympy.Eq(Gamma[a][b][c], 0).simplify()


def test_angular_christoffels(angular_model):
    Gamma = angular_model.christoffels()
    for a in range(angular_model.dim):
        for b in range(angular_model.dim):
            for c in range(angular_model.dim):
                assert sympy.Eq(Gamma[a][b][c], Gamma[a][c][b]).simplify()


def test_gramm_schmidt(trivial_model):
    a, b = sympy.symbols("a b")
    v1 = trivial_model.normalize([1, a**2])
    v2 = [sympy.sqrt(b), sympy.sin(a)]
    v2_ortho = trivial_model.gramm_schmidt([v1], v2)
    assert sympy.Eq(trivial_model.inner_prod(v2_ortho, v2_ortho), 1).simplify()
    assert sympy.Eq(trivial_model.inner_prod(v1, v2_ortho).simplify(), 0).simplify()



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# System imports
import os
import subprocess
import sys
import tempfile
from datetime import datetime
from sys import version as sys_version

# Sympy imports
import sympy
from sympy.printing.c import C99CodePrinter

# Internal imports
from .symbolic import InflationModel
from .version import __abi_version__, __version__


class CInflatoxPrinter(C99CodePrinter):
    """C99CodePrinter with modified `_print_Symbol` method. Converting Sympy
    expressions with this printer will map all sympy symbols to either:
      - `x[i]` for symbols that must be interpreted as coordinates on the scalar
      manifold
      - `args[i]` for other symbols
    Which symbols should be interpreted as coordinates and which ones should not
    can be specified with the class constructor by passing it the appropriate
    value for the `coordinate_symbols` argument.
    """

    def __init__(
        self,
        coordinate_symbols: list[sympy.Symbol],
        coordinate_derivative_symbols: list[sympy.Symbol],
        settings=None,
    ):
        super().__init__(settings)
        coord_dict = {}
        for i, symbol in enumerate(coordinate_symbols):
            coord_dict[super()._print_Symbol(symbol)] = f"x[{i}]"
        dotcoord_dict = {}
        for i, symbol in enumerate(coordinate_derivative_symbols):
            dotcoord_dict[super()._print_Symbol(symbol)] = f"xdot[{i}]"
        self.coord_dict = coord_dict
        self.dotcoord_dict = dotcoord_dict
        self.param_dict = {}

    def print_preamble(self, model_name: str):
        """prints preamble in generated c code"""
        return f"""// This source file was automatically generated by Inflatox
// Model: {model_name}, timestamp: {datetime.now().strftime("%Y-%m-%d, %H:%M:%S")}
// Inflatox version: v{__version__}
// System info: {sys_version}

#include <math.h>
#include <stdint.h>
#ifndef M_PI
#define M_E 2.71828182846 
#define M_LOG2E 1.44269504089
#define M_LOG10E 0.4342944819
#define M_LN2 0.69314718056
#define M_LN10 2.30258509299
#define M_PI 3.14159265359
#define M_PI_2 1.57079632679
#define M_PI_4 0.78539816339
#define M_1_PI 0.31830988618
#define M_2_PI 0.63661977236
#define M_2_SQRTPI 1.1283791671
#define M_SQRT2 1.41421356237
#define M_SQRT_1_2 0.70710678118
#endif
"""

    def _print_Symbol(self, expr):
        """Modified _print_Symbol function that maps sympy symbols to argument indices"""
        if expr.is_number:
            return expr.evalf(self._settings["precision"])

        sym_name = self.get_symbol(expr)
        if sym_name is not None:
            return sym_name
        else:
            return self.register_parameter(expr)

    def register_parameter(self, symbol: sympy.Symbol) -> str:
        """Adds symbol to the parameter dictionary"""
        sym_name = f"args[{len(self.param_dict)}]"
        self.param_dict[super()._print_Symbol(symbol)] = sym_name
        return sym_name

    def get_symbol(self, symbol: sympy.Symbol) -> str | None:
        """Returns string representing sympy symbol"""
        sym_name = super()._print_Symbol(symbol)
        if sym_name.startswith("cse"):
            return sym_name
        if self.coord_dict.get(sym_name) is not None:
            return self.coord_dict[sym_name]
        elif self.dotcoord_dict.get(sym_name) is not None:
            return self.dotcoord_dict[sym_name]
        elif self.param_dict.get(sym_name) is not None:
            return self.param_dict[sym_name]
        else:
            return None


class GSLInflatoxPrinter(CInflatoxPrinter):
    """An extended version of the `CInflatoxPrinter` capable of converting some special functions to
    their GSL counterparts. For a list of supported functions, see `Compiler` class documentation."""

    HYPERH = "gsl_sf_hyperg"
    BESSELH = "gsl_sf_bessel"

    def __init__(
        self,
        coordinate_symbols: list[sympy.Symbol],
        coordinate_derivative_symbols: list[sympy.Symbol],
        settings=None,
    ):
        super().__init__(coordinate_symbols, coordinate_derivative_symbols, settings)
        self.required_headers = []

    def print_preamble(self, model_name: str):
        """Print necessary gsl include directives, as well as the err_setup() function, called by
        lib_inflx_rs when the library is loaded."""
        preamble = super().print_preamble(model_name)
        for required_header in self.required_headers:
            preamble += f"#include<gsl/{required_header}.h>"
        preamble += """// Declare external function for gsl error handling
#include<gsl/gsl_errno.h>
void err_setup(gsl_error_handler_t* rust_panic) {
    gsl_set_error_handler(rust_panic);
}
    """
        return preamble

    def update_preamble(self, header):
        """Add header to preamble"""
        if header not in self.required_headers:
            self.required_headers.append(header)

    def _print_hyper(self, expr):
        """printer for hypergeometric functions"""
        self.update_preamble(self.HYPERH)
        ap = expr.args[0]
        bq = expr.args[1]
        x = self.doprint(expr.args[2])

        type = (len(ap), len(bq))
        if type == (2, 0):
            return f"gsl_sf_hyperg_2F0({self.doprint(ap[0])}, {self.doprint(ap[1])}, {x})"
        elif type == (2, 1):
            return f"gsl_sf_hyperg_2F1({self.doprint(ap[0])}, {self.doprint(ap[1])}, {self.doprint(bq[0])}, {x})"
        elif type == (1, 1):
            return f"gsl_sf_hyperg_1F1({self.doprint(ap[0])}, {self.doprint(bq[0])}, {x})"
        elif type == (0, 1):
            return f"gsl_sf_hyperg_0F1({self.doprint(bq[0])}, {x})"
        else:
            raise Exception(
                "Cannot compute hypergeometric functions other than 2F0, 2F1, 1F1 and 0F1"
            )

    def generic_print_bessel(self, expr, namedict):
        """printer for Bessel family of functions"""
        self.update_preamble(self.BESSELH)
        nu = expr.args[0]
        x = self._print_Symbol(expr.args[1])
        if nu.is_integer:
            nu = int(float(self._print_Symbol(nu)))
            try:
                name = namedict[f"{nu}"]
                return f"gsl_sf_bessel_{name}({x})"
            except KeyError:
                return f"gsl_sf_bessel_{namedict['int']}({nu}, {x})"
        try:
            return f"gsl_sf_bessel_{namedict['float']}({self._print_Symbol(nu)}, {x})"
        except KeyError:
            raise KeyError("No non-integer impl found.")

    def _print_besselj(self, expr):
        return self.generic_print_bessel(expr, {"0": "J0", "1": "J1", "int": "Jn", "float": "Jnu"})

    def _print_bessely(self, expr):
        return self.generic_print_bessel(expr, {"0": "Y0", "1": "Y1", "int": "Yn", "float": "Ynu"})

    def _print_besseli(self, expr):
        return self.generic_print_bessel(expr, {"0": "I0", "1": "I1", "int": "In", "float": "Inu"})

    def _print_besselk(self, expr):
        return self.generic_print_bessel(expr, {"0": "K0", "1": "K1", "int": "Kn", "float": "Knu"})

    def _print_jn(self, expr):
        return self.generic_print_bessel(expr, {"0": "j0", "1": "j1", "2": "j2", "int": "jl"})

    def _print_yn(self, expr):
        return self.generic_print_bessel(expr, {"0": "y0", "1": "y1", "2": "y2", "int": "yl"})


class CompilationArtifact:
    """Class representing the output of the `Compiler`. It contains all information
    necessary to access the compiled artifact.

    ### Compiler symbols
    The `Compiler` class maps all sympy symbols found in the expressions for the
    potential and projected Hesse matrix to arguments of two numpy arrays:
      - `x` for the scalar fields themselves.
      - `args` for all other symbols (model parameters).
    All functions and classes that make use of this `CompilationArtifact` class will
    most likely require the user to supply `x` and `args` as numpy arrays. Therefore,
    one must know which sympy symbols were mapped to which position in the `x` and
    `args` arrays. The `CompilationArtifact` class provides two methods for this:
    `symbol_lookup` and `print_sym_table`. See their documentation for more details.
    """

    symbol_printer = C99CodePrinter()

    def __init__(
        self,
        symbol_dictionary: dict,
        shared_object_path: str,
        n_fields: int,
        n_parameters: int,
        auto_cleanup: bool = True,
    ):
        self.symbol_dictionary = symbol_dictionary
        self.shared_object_path = shared_object_path
        self.n_fields = n_fields
        self.n_parameters = n_parameters
        self.auto_cleanup = auto_cleanup

    def __del__(self):
        # Delete compilation artifact
        if self.auto_cleanup:
            os.remove(self.shared_object_path)

    def lookup_symbol(self, symbol: sympy.Symbol) -> str | None:
        """returns the compiled symbol (string) for the supplied sympy symbol,
        if the sympy symbol is known, `None` otherwise. See class docs for more
        info on compiled symbols.

        ### Args
        `symbol` (`sympy.Symbol`): sympy symbol to be converted.

        ### Returns
        `str|None`: name of the supplied symbol (either `args[n]` or `x[n]`), or
        `None` if the symbol is unknown.
        """
        sym_name = self.symbol_printer._print_Symbol(symbol)
        if not isinstance(sym_name, str):
            return None
        else:
            return self.symbol_dictionary[sym_name]

    def print_sym_lookup_table(self):
        """prints full mapping of all known sympy symbols and their corresponding
        compiled symbols. See class docs for more info on compiled symbols.
        """
        print("[Symbol Dictionary]")
        for old, new in self.symbol_dictionary.items():
            print(f"{old} -> {new}")


class Compiler:
    """This class wraps the native platform C compiler. It can be used to generate,
    compile and link C code from a `HesseMatrix` instance to produce a `CompilationArtifact`
    which can be used to calculate consistency conditions. This process involves
    creating a symbol dictionary that maps all symbols used in the `HesseMatrix` to
    C-friendly symbols.

    ## Special function support
    By passing `link_gsl=True` to the constructor of this class, the GSL will be linked by the final
    binary. This allows inflatox to map some special functions from `scipy` to their GSL counterparts.
    Currently supported functions are:
    - Bessel functions (besselj, besseli, besselk, bessely, jn and yn)
    - Hypergeometric functions (2F1, 2F0, 1F1 and 2F0)
    If you have the need for more special functions or are experiencing other issues with the gsl
    feature, contact the authors or open an issue on [github](https://github.com/smups/inflatox/issues)
    """

    c_prefix = "inflx_auto_"
    lib_prefix = "libinflx_auto_"

    default_zigcc_flags = [
        "-O3",
        "-Wall",
        "-Werror",
        "-fpic",
        "-lm",
        "-march=native",
        "-shared",
        "-std=c17",
        "-fno-math-errno",
        "-fno-signed-zeros",
    ]
    # The linker option -Wl --no-as-needed is required on Fedora/Redhat based systems. It could
    # very well be the case that this does not work on Debian/Ubuntu based systems.
    gsl_zigcc_flags = ["-lgsl", "-Wl,--no-as-needed", "-lgslcblas"]

    def __init__(
        self,
        model: InflationModel,
        output_path: str | None = None,
        cleanup: bool = True,
        silent: bool = False,
        link_gsl: bool = False,
        cse: bool = False,
        max_cses: int = 1000,
        compiler_flags: list[str] | None = None,
    ):
        """Constructor for a C Compiler (provided by zig-cc), which can be used
        to convert the provided `HesseMatrix` object into a platform- and arch-specific
        shared library object.

        ### Examples
        To compile a previously calculated Hesse matrix, we simply construct a
        `Compiler` instance and call `.compile()` on it:
        ```python
        artifact = inflatox.Compiler(hesse_matrix).compile()
        ```
        See the docs for `HesseMatrix` and `SymbolicCalculation` for info on how to
        obtain a `HesseMatrix` instance.

        ### Args
        - `hesse_matrix` (HesseMatrix): HesseMatrix object that will be turned into
          C code and compiled.
        - `output_path` (str | None, optional): output path of compilation artifacts.
          Will auto-select the platform-defined temporary folder if option is set to
          `None`. Defaults to `None`.
        - `cleanup` (bool, optional): if `True`, generated artifacts will be deleted
          when they are no longer necessary. Defaults to True.
        - `silent` (bool, optional): if `True`, no console output will be
          generated. Defaults to `False`
        - `link_gsl` (bool, optional): enables experimental gsl (GNU Scientific Library) support for
          output binary. This enables some special functions to be compiled by inflatox. Enabling this
          binary requires the gls library to be installed and available to the linker. Defaults to False.
        - `cse` (bool, optional): if `True`, common subexpressions will be split into variables.
          This enables additional optimizations by the compiler, BUT MAY BREAK CODE! Defaults to False.
        - `max_cses` (int, optional): maximum number of common subexpressions within a single function.
           This limit is set primarily to avoid hitting an infinite loop when translating sympy
           expressions to C functions using common subexpression elimination. Defaults to 1000.
        - `compiler_flags` (list(str)|None, optional): replace default compiler flags with user-supplied
          ones. Make sure to link libmath and libgsl/libgslcblas (if link_gls==True). The default compiler
          flags can be found under `Compiler.default_zigcc_flags` and `Compiler.gsl_zigcc_flags`.
          Defaults to None.
        """
        self.output_file = (
            open(output_path, "w")
            if output_path is not None
            else tempfile.NamedTemporaryFile(
                mode="wt", delete=False, suffix=".c", prefix=Compiler.c_prefix
            )
        )
        self.symbolic_out = model
        self.cleanup = cleanup
        self.silent = silent
        self.gsl = link_gsl
        self.cse = cse
        self.max_cses = max_cses
        self.zigcc_opts = [flag for flag in self.default_zigcc_flags]
        # Add gsl linker flags
        if link_gsl:
            for flag in self.gsl_zigcc_flags:
                self.zigcc_opts.append(flag)
        # Override default flags if user supplied some
        if compiler_flags is not None:
            self.zigcc_opts = compiler_flags

    def _new_cse_generator(self):
        """Returns a generator that yields new symbols for common subexpressions until the
        user-specified maximum is reached.
        """

        def symbol_generator():
            num = 0
            while num <= self.max_cses:
                yield sympy.symbols(f"cse{num}")
                num += 1
            raise Exception("Maximum number of common subexpressions reached!")

        return symbol_generator()

    def _generate_c_function(
        self, fn_signature: str, body: sympy.Expr, printer: CInflatoxPrinter
    ) -> str:
        out = fn_signature + "{\n"
        """Generates a function body from the provided expression"""
        if self.cse:
            cse_list = sympy.cse(body, symbols=self._new_cse_generator(), order="none", list=False)
            if not self.silent:
                print(f"Found {len(cse_list[0])} common subexpressions")
            for cse_symbol, cse_definition in cse_list[0]:
                out += f"    const double {printer.doprint(cse_symbol)} = {printer.doprint(cse_definition)};"
                out += "\n"
            out += f"    return {printer.doprint(cse_list[1])};"
            out += "\n}\n"
        else:
            out += f"    return {printer.doprint(body)};"
            out += "\n}\n"
        return out + "\n"

    def _generate_c_function_for_vector(
        self, fn_signature: str, vector: list[sympy.Expr], printer: CInflatoxPrinter
    ) -> str:
        """Generates a function body for a vector expression"""
        out = fn_signature + "{\n"
        ordered_output_expr = []

        # First print subexpressions (possibly common to *all* components of the vector)
        if self.cse:
            cse_list = sympy.cse(vector, symbols=self._new_cse_generator(), list=True)
            if not self.silent:
                print(f"Found {len(cse_list[0])} common subexpressions")
            for cse_symbol, cse_definition in cse_list[0]:
                out += f"    const double {printer.doprint(cse_symbol)} = {printer.doprint(cse_definition)};"
                out += "\n"
            for output_expr in cse_list[1]:
                ordered_output_expr.append(output_expr)
        else:
            ordered_output_expr = vector

        # Assign each element of the output vector to a component of the vector
        for i, output_cmp in enumerate(ordered_output_expr):
            out += f"    v_out[{i}] = {printer.doprint(output_cmp)};"
            out += "\n"
        out += "    return;\n}\n\n"

        return out

    def _generate_c_function_for_inner_prod(self, printer: CInflatoxPrinter) -> str:
        out = "double inner_prod(const double x[], const double args[], const double v1[], const double v2[]){\n"
        flattened_metric = []
        for i in range(self.symbolic_out.dim):
            for j in range(self.symbolic_out.dim):
                flattened_metric.append(self.symbolic_out.metric[i][j])
        if self.cse:
            cse_list = sympy.cse(flattened_metric, symbols=self._new_cse_generator(), list=True)
            for cse_symbol, cse_definition in cse_list[0]:
                out += f"    const double {printer.doprint(cse_symbol)} = {printer.doprint(cse_definition)};"
                out += "\n"
            flattened_metric = cse_list[1]

        # Write function for outer product
        return_expr = "0.0"
        for i in range(self.symbolic_out.dim):
            n = self.symbolic_out.dim * i
            for j in range(self.symbolic_out.dim):
                n += j
                symbol_str = printer.doprint(flattened_metric[n])
                if symbol_str == "0" or symbol_str == "0.0":
                    continue
                out += f"    const double g{i}{j} = {symbol_str};"
                out += "\n"
                return_expr += f" + (g{i}{j} * v1[{i}] * v2[{j}])"
        out += f"    return {return_expr};"
        out += "\n}\n\n"
        return out

    def _generate_c_file(self):
        """Generates C source file from Hesse matrix specified by the constructor"""
        # Initialise C-code printer
        fields = self.symbolic_out.coordinates
        dotfields = self.symbolic_out.coordinate_tangents
        printer = (
            GSLInflatoxPrinter(fields, dotfields)
            if self.gsl
            else CInflatoxPrinter(fields, dotfields)
        )

        if not self.silent and self.cse:
            print("Converting sympy to C using common subexpression elimination...")
        contents = ""

        # Write potential
        contents += self._generate_c_function(
            "double V(const double x[], const double args[])", self.symbolic_out.potential, printer
        )

        # Use metric to create function for evaluating inner prods.
        contents += self._generate_c_function_for_inner_prod(printer)

        # Write all the components of the Hesse matrix
        for a in range(self.symbolic_out.dim):
            for b in range(self.symbolic_out.dim):
                contents += self._generate_c_function(
                    f"double v{a}{b}(const double x[], const double args[])",
                    self.symbolic_out.hesse_cmp[a][b],
                    printer,
                )

        # Output functions for each of the basis vectors
        for idx in range(self.symbolic_out.dim):
            vector = self.symbolic_out.basis[idx]
            print(vector)
            name = "v" if idx == 0 else f"w{idx}"
            signature = f"void {name}(const double x[], const double args[], double v_out[])"
            contents += self._generate_c_function_for_vector(signature, vector, printer)

        # Write the size of the gradient
        contents += self._generate_c_function(
            "double grad_norm_squared(const double x[], const double args[])",
            self.symbolic_out.gradient_square,
            printer,
        )

        # Write the equations of motion
        for a in range(self.symbolic_out.dim):
            contents += self._generate_c_function(
                f"double eom{a}(const double x[], const double xdot[], const double args[])",
                self.symbolic_out.eom_fields[a],
                printer,
            )

        # Write the equations of motion for the hubble parameter
        contents += self._generate_c_function(
            "double eomh(const double x[], const double xdot[], const double args[])",
            self.symbolic_out.eom_h,
            printer,
        )
        contents += self._generate_c_function(
            "double eomhdot(const double x[], const double xdot[], const double args[])",
            self.symbolic_out.eom_hdot,
            printer,
        )

        # Output to actual file
        with self.output_file as out:
            # Write preamble
            out.write(printer.print_preamble(self.symbolic_out.model_name))

            # Write global constants
            v = __abi_version__.split(".")
            out.write(f"""
//Inflatox version used to generate this file
const uint16_t VERSION[3] = {{{v[0]},{v[1]},{v[2]}}};
//Number of fields (dimensionality of the scalar manifold)
const uint32_t DIM = {self.symbolic_out.dim};
//Number of parameters
const uint32_t N_PARAMETERS = {len(printer.param_dict)};
//Model name
char *const MODEL_NAME = \"{self.symbolic_out.model_name}\";
// Gsl flag
const char USE_GSL = {1 if self.gsl else 0};

""")
            # Write actual file contents
            out.write(contents)

        # Update symbol dictionary
        self.symbol_dict = printer.coord_dict
        self.symbol_dict.update(printer.param_dict)

    def _zigcc_compile_and_link(self):
        source_path = f"{self.output_file.name}"
        source_name = os.path.basename(source_path)[:-2].removeprefix(Compiler.c_prefix)
        lib_name = f"{Compiler.lib_prefix}{source_name}.bin"
        lib_path = f"{tempfile.tempdir}/{lib_name}"

        # Compile source with zig
        zigargs = [
            sys.executable,
            "-m",
            "ziglang",
            "cc",
            "-o",
            lib_path,  # out
            source_path,  # in
            *self.zigcc_opts,  # compiler options
        ]

        process = subprocess.Popen(zigargs, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out = b""

        while process.stdout.readable():
            line = process.stderr.readline()
            if not line:
                break
            out += line
            if not self.silent:
                print(line.decode("utf-8"), end=None)

        exitcode = process.wait()
        return (source_path, lib_path, (out, exitcode))

    def compile(self) -> CompilationArtifact:
        """Compiles the Hesse matrix specified in the constructor of this class into
        a shared library that can be used in further calculations. This process involves
        creating a symbol dictionary that maps all symbols used in the `HesseMatrix`
        to C-friendly symbols. The output of this function contains this dictionary,
        see the docs of `CompilationArtifact` for more info.

        ### Examples
        To compile a previously calculated Hesse matrix, we simply construct a
        `Compiler` instance and call `.compile()` on it:
        ```python
        artifact = inflatox.Compiler(hesse_matrix).compile()
        ```
        See the docs for `HesseMatrix` and `SymbolicCaluclation` for info on how to
        obtain a `HesseMatrix` instance.

        ### Returns
          `CompilationArtifact`: artifact that can be used in further calculations.
          It contains info about the model and inflatox version used to create the
          artifact.
        """
        # (0) Say hello
        if not self.silent:
            print("Compiling model...")

        # (1) generate the actual C-source
        self._generate_c_file()

        # (2) run compiler and linker
        source_path, dylib_path, (output, exitcode) = self._zigcc_compile_and_link()

        # (3) cleanup unused artifacts
        if self.cleanup:
            os.remove(source_path)

        # (4) print output
        if exitcode != 0:
            if self.silent:
                print(output.decode("utf-8"))

            print(f'Problematic source file located at: "{source_path}"')
            raise Exception("Zig compiler error (see previous output)")

        # (R) return compilation artifact
        return CompilationArtifact(
            self.symbol_dict,
            dylib_path,
            self.symbolic_out.dim,
            len(self.symbol_dict) - self.symbolic_out.dim,
            auto_cleanup=self.cleanup,
        )



#  Copyright© 2024 Raúl Wolters(1)
#
#  This file is part of Inflatox.
#
#  Inflatox is free software: you can redistribute it and/or modify it under
#  the terms of the European Union Public License version 1.2 or later, as
#  published by the European Commission.
#
#  Inflatox is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the European Union Public License for more details.
#
#  You should have received a copy of the EUPL in an/all official language(s) of
#  the European Union along with Inflatox.  If not, see
#  <https://ec.europa.eu/info/european-union-public-licence_en/>.
#
#  (1) Resident of the Kingdom of the Netherlands; agreement between licensor and
#  licensee subject to Dutch law as per article 15 of the EUPL.

# ************************************************************************************************ #
# Angular Inflation used in Wolters, Iarygina and Achúcarro JCAP07(2024)079
# Model from Christodoudilis, Roest and Sfakianakis JCAP11(2019)002
# ************************************************************************************************ #

import os

import inflatox
import numpy as np
import sympy
from inflatox.consistency_conditions import GeneralisedAL

trajectory_dir = f"{os.path.dirname(os.path.abspath(__file__))}/trajectories/"


def test_angular():
    model = "angular"

    # setup the coordinates
    p, x = sympy.symbols("phi chi")
    coords = [p, x]

    # setup the potential
    mp, mx, a = sympy.symbols("m_phi m_chi alpha")
    potential = a / 2 * ((mp * p) ** 2 + (mx * x) ** 2).nsimplify()

    # setup the metric
    metric_diagonal = 6 * a / (1 - p**2 - x**2) ** 2
    metric = [[0 for _ in range(2)] for _ in range(2)]
    metric[0][0] = metric_diagonal
    metric[1][1] = metric_diagonal

    hesse = inflatox.InflationModelBuilder.new(
        coords,
        metric,
        potential,
        model_name=model,
        silent=True,
    ).build()

    out = inflatox.Compiler(hesse, cleanup=False, cse=True).compile()
    anguelova = GeneralisedAL(out)

    a = 1 / 600
    m_phi = 2e-5
    m_chi = m_phi * np.sqrt(9)
    args = np.array([a, m_chi, m_phi])

    extent = (-1.05, 1.05, -1.05, 1.05)
    phi_start, phi_stop = -15.0, 15.0
    chi_start, chi_stop = -5.0, 5.0
    N = 100

    # Calculate potential
    anguelova.calc_V_array(args, [phi_start, chi_start], [phi_stop, chi_stop], [N, N])

    # run analysis
    anguelova.complete_analysis(args, *extent, *[N, N])

    # run analysis on trajectory
    tx = np.load(f"{trajectory_dir}/angular_phix.npy")
    ty = np.load(f"{trajectory_dir}/angular_phiy.npy")
    trajectory = np.column_stack((tx, ty))
    anguelova.complete_analysis_ot(args, trajectory)

    # run Anguelova's original condition
    anguelova.consistency_rapidturn(args, *extent, *[N, N])
